// https://observablehq.com/@norswap/fitbit/2@1086
export default function define(runtime, observer) {
  const main = runtime.module();
  main.variable(observer()).define(["md"], function(md){return(
md`# Fitbit`
)});
  main.variable(observer()).define(["md"], function(md){return(
md`
This notebook contains some simple logic to export daily Fitbit data (namely, total calories in, total calories out, and weight) to an (empty) Google Sheet spreadsheet.

See the "Setup" section below for instructions.
`
)});
  main.variable(observer()).define(["md"], function(md){return(
md `# Triggers`
)});
  main.variable(observer("viewof google_signin_status")).define("viewof google_signin_status", ["html"], function(html){return(
html `<p></p>`
)});
  main.variable(observer("google_signin_status")).define("google_signin_status", ["Generators", "viewof google_signin_status"], (G, _) => G.input(_));
  main.variable(observer("viewof fitbit_signin_status")).define("viewof fitbit_signin_status", ["html"], function(html){return(
html `<p></p>`
)});
  main.variable(observer("fitbit_signin_status")).define("fitbit_signin_status", ["Generators", "viewof fitbit_signin_status"], (G, _) => G.input(_));
  main.variable(observer("viewof write_sheet_button")).define("viewof write_sheet_button", ["html","write_sheet","swal"], function(html,write_sheet,swal)
{
  const root = html `<button>Write Sheet!</button>`
  root.onclick = () => {
    write_sheet()
    swal("Success", "Sheet written.", "success");
  }
  return root
}
);
  main.variable(observer("write_sheet_button")).define("write_sheet_button", ["Generators", "viewof write_sheet_button"], (G, _) => G.input(_));
  main.variable(observer()).define(["md"], function(md){return(
md `# Config`
)});
  main.variable(observer("GAPI_KEY")).define("GAPI_KEY", ["Secret"], function(Secret){return(
Secret("GAPI_KEY")
)});
  main.variable(observer("GCLIENT_ID")).define("GCLIENT_ID", ["Secret"], function(Secret){return(
Secret("GCLIENT_ID")
)});
  main.variable(observer("FITBIT_OAUTH_ID")).define("FITBIT_OAUTH_ID", ["Secret"], function(Secret){return(
Secret("FITBIT_OAUTH_ID")
)});
  main.variable(observer("viewof OAUTH_URL")).define("viewof OAUTH_URL", ["md","html","FITBIT_OAUTH_ID"], function(md,html,FITBIT_OAUTH_ID)
{
  const root =  md `\`OAUTH_URL\``
  const this_page = encodeURIComponent(html`<a href>`.href)
  root.value = `https://www.fitbit.com/oauth2/authorize?response_type=token&client_id=${FITBIT_OAUTH_ID}&redirect_uri=${this_page}&scope=activity%20heartrate%20location%20nutrition%20profile%20settings%20sleep%20social%20weight&expires_in=604800`
  return root 
}
);
  main.variable(observer("OAUTH_URL")).define("OAUTH_URL", ["Generators", "viewof OAUTH_URL"], (G, _) => G.input(_));
  main.variable(observer("SPREADSHEET_ID")).define("SPREADSHEET_ID", function(){return(
"1fwz-pYvr_x543N_b27kdIKfXY0T7ACZgl7V3UgoqNBE"
)});
  main.variable(observer("SHEET_NAME")).define("SHEET_NAME", function(){return(
"Sheet1"
)});
  main.variable(observer("START_DATE")).define("START_DATE", function(){return(
"2019-04-22"
)});
  main.variable(observer("END_DATE")).define("END_DATE", function(){return(
"2019-05-22"
)});
  main.variable(observer()).define(["md"], function(md){return(
md `# Setup

## Fitbit

- Create a Fitbit dev account.

- Go to https://dev.fitbit.com/apps and create a new app.
  - The type of application should be "client"
  - The callback URL should be the URL of your work of this notebook.
  - Read-only access is enough.

- In your ObservableHQ settings, create a secret calling \`FITBIT_OAUTH_ID\` and fill it with the "OAuth 2.0 Client ID" for your app.

## Google Sheets

- Go to https://console.developers.google.com/apis/ and enable the Google Sheets API.

- Go to https://console.developers.google.com/apis/credentials and create new OAuth client ID credentials.

- In your ObservableHQ settings, create a secret calling \`GCLIENT_ID\` and fill it with the client ID you just created.

- On the Google developer console, create a new API key. You can set restrictions if you wish (HTTP referrers / Google Sheets API).

- In your ObservableHQ settings, create a secret calling \`GAPI_KEY\` and fill it with the API key you just created.

- In Google Sheets, create a new spreadsheet and copy its [Spreadsheet ID](https://developers.google.com/sheets/api/guides/concepts/#spreadsheet_id) (which appears in the URL) — copy it into the \`SPREADSHEET_ID\` variable of this notebook.

## Configure The Notebook & Export

- Set the \`START_DATE\` and \`END_DATE\` variables of this notebook to the desired start/end dates for the data you wish to export to Google Sheets. Currently, these two dates must not be more than 31 days apart. A change to the \`weight_range\` function could fix this, by splitting a bigger range into at-most-31-days chunks, making multiple API request and stitching their results together.

- Sign in to Google using the button in the "Trigger" section.

- Sign in to Fitbit using the button in the "Trigger" section (this will reload the page and append an access token to the URL).

- Press the "Write Sheet!" button to export data to your spreadsheet.
  - If successful, a popup will appear.
  - Note that currently, errors are only logged in the Javascript developer console.

`
)});
  main.variable(observer()).define(["md"], function(md){return(
md `# Implem`
)});
  main.variable(observer("write_sheet")).define("write_sheet", ["SHEET_NAME","parse_date","START_DATE","END_DATE","calories_in_range","calories_out_range","weight_range","awaitify_req","gapi","SPREADSHEET_ID"], function(SHEET_NAME,parse_date,START_DATE,END_DATE,calories_in_range,calories_out_range,weight_range,awaitify_req,gapi,SPREADSHEET_ID){return(
async () =>
{
  const first_column  = `${SHEET_NAME}!A:F`
  const start_date    = parse_date(START_DATE)
  const end_date      = parse_date(END_DATE)
  
  if (!start_date.isBefore(end_date))
    throw "Non sequential request date range."
  
  const calories_in  = await calories_in_range(START_DATE, END_DATE)
  const calories_out = await calories_out_range(START_DATE, END_DATE)
  const weight       = await weight_range(START_DATE, END_DATE)
  
  const value_map = {}
  
  for (const it of calories_in)
    value_map[it.dateTime] = { calories_in: it.value }
  
  for (const it of calories_out)
    value_map[it.dateTime].calories_out = it.value
  
  for (const it of weight)
    value_map[it.date].weight = it.weight
  
  const values = []
  
  for (const date in value_map) {
    const data = value_map[date] 
    values.push([date, data.calories_in, data.calories_out, data.weight])
  }
  
  const style_requests = []
  
  const BLACK = {
    red:   0,
    green: 0,
    blue:  0,
    alpha: 1
  }
  
  for (let i = 0; i < values.length; ++i)
  {
    const row = values[i]
    row.push(`=B${i+1}-C${i+1}`) // diff = calories_in - calories_out
    
    const floor = Math.max(1, i-1)
    row.push(`=average(D${floor}:D${i+3})`) // rolling smoothed average over 5 cells
    
    // black medium borders below each sunday row
    // NOTE: these borders are never cleared when the sheet is updated!
    if (parse_date(row[0]).day() == 0)
      style_requests.push({
        updateBorders: {
          range: {
            sheetId: 0,
            startRowIndex: i,
            endRowIndex: i+1
          },
          bottom: {
            style: "SOLID_MEDIUM",
            color: BLACK
          }
        }
      })
  }
  
  const r1 = await awaitify_req(gapi.client.sheets.spreadsheets.values.update({
    spreadsheetId: SPREADSHEET_ID,
    range: first_column,
    valueInputOption: "USER_ENTERED",
    resource: {
      majorDimension: "ROWS",
      values: values }
  }))
  
  // console.log(r1)
  
  const r2 = await awaitify_req(gapi.client.sheets.spreadsheets.batchUpdate({
    spreadsheetId: SPREADSHEET_ID,
    resource: {requests: style_requests}
  }))
  
  // console.log(r2)
}
)});
  main.variable(observer()).define(["md"], function(md){return(
md`## Signup Logic`
)});
  main.variable(observer("viewof gapi")).define("viewof gapi", ["html","viewof google_signin_status","require","GAPI_KEY","GCLIENT_ID","md"], async function(html,$0,require,GAPI_KEY,GCLIENT_ID,md)
{
  // https://developers.google.com/sheets/api/quickstart/js
  
  function google_signin() {
    gapi.auth2.getAuthInstance().signIn();
  }
  
  function google_signout() {
    gapi.auth2.getAuthInstance().signOut();
  }
  
  function update_signin_status (signed_in) {
    // https://talk.observablehq.com/t/d3-to-populate-buttons-w-onclick-function/1640
    if (signed_in) {
      const button = html `<button>Sign out</button>`
      button.onclick = google_signout
      $0.innerHTML = "Signed into Google"
      $0.appendChild(html`&emsp;`)
      $0.appendChild(button)
    } else {
      const button = html `<button>Click to sign into Google</button>`
      button.onclick = google_signin
      $0.innerHTML = ''
      $0.appendChild(button)
    }
  }
  
  // https://talk.observablehq.com/t/importing-googleapis/2190
  const gapi = await require("https://apis.google.com/js/api.js").catch(() => window.gapi)
  await new Promise((ok, err) => gapi.load("client:auth2", {callback: ok, onerror: err}))
  
  // must whitelist https://norswap.static.observableusercontent.com/ on google dev console
  await gapi.client.init({
    apiKey: GAPI_KEY,
    clientId: GCLIENT_ID,
    discoveryDocs: ["https://sheets.googleapis.com/$discovery/rest?version=v4"],
    // https://developers.google.com/identity/protocols/googlescopes#sheetsv4
    scope: "https://www.googleapis.com/auth/spreadsheets"
  })
  .catch(e => console.log(`Error while initializing the Google Client: ${JSON.stringify(e, null, 2)}`))
  
  // Listen for sign-in state changes.
  gapi.auth2.getAuthInstance().isSignedIn.listen(update_signin_status);
  
  // Handle the initial sign-in state.
  update_signin_status(gapi.auth2.getAuthInstance().isSignedIn.get());
  
  const root = md `\`gapi\``
  root.value = gapi
  return root
}
);
  main.variable(observer("gapi")).define("gapi", ["Generators", "viewof gapi"], (G, _) => G.input(_));
  main.variable(observer("viewof fitbit_iden")).define("viewof fitbit_iden", ["html","OAUTH_URL","viewof fitbit_signin_status","location","md"], function(html,OAUTH_URL,$0,location,md)
{
  
  if (window.location.hash == "") {
    const button = html `<button><a href="${OAUTH_URL}">Click to sign into Fitbit</a></button>`
    button.children[0].style.textDecoration = "none"
    button.children[0].style.color = "black"
    $0.innerHTML = ''
    $0.appendChild(button)
    const root = html `<code>fitbit_iden = <span class="observablehq--red">false</span></code>`
    root.value = false
    return root
  }
  
  function fitbit_signout (e) {
    console.log("prout")
    window.location.hash = ""
    e.preventDefault()
    console.log(window.location.hash)
    location.reload(true)
  }
  
  // There is no easy way to "sign out" (remove hashes from URL (easy) + ensure the address bar is updated (hard))
  $0.innerHTML = "Signed into Fitbit"
  const params = {}
  window.location.hash.substring(1).split('&').map(hk => { 
    const pair = hk.split('=');
    params[pair[0]] = pair[1] 
  })
  
  const root = md `\`fitbit_iden\``
  root.value = params
  return root
}
);
  main.variable(observer("fitbit_iden")).define("fitbit_iden", ["Generators", "viewof fitbit_iden"], (G, _) => G.input(_));
  main.variable(observer()).define(["md"], function(md){return(
md `## Helpers`
)});
  main.variable(observer("moment")).define("moment", ["require"], function(require){return(
require("moment")
)});
  main.variable(observer("format_date")).define("format_date", function(){return(
date => date.format("YYYY-MM-DD")
)});
  main.variable(observer("parse_date")).define("parse_date", ["moment"], function(moment){return(
string => moment(string, "YYYY-MM-DD")
)});
  main.variable(observer("promisify")).define("promisify", function(){return(
thenable => new Promise((resolve, reject) => thenable.then(resolve, reject))
)});
  main.variable(observer("awaitify_req")).define("awaitify_req", ["promisify"], function(promisify){return(
thenable => promisify(thenable).catch(error => { throw error.body })
)});
  main.variable(observer("swal")).define("swal", ["require"], function(require){return(
require("sweetalert")
)});
  main.variable(observer()).define(["md"], function(md){return(
md `## Fitbit API Calls`
)});
  main.variable(observer("fitbit_range_call")).define("fitbit_range_call", ["fitbit_iden"], function(fitbit_iden){return(
async (endpoint, access_key) => {
  
  const response = await fetch(endpoint, { headers: {
      "Authorization": `${fitbit_iden.token_type} ${fitbit_iden.access_token}`
    }}).catch(err => { throw err })
  
  const json = await response.json()
  
  if (json.errors)
    throw json.errors[0].message
  
  return json[access_key]
}
)});
  main.variable(observer("calories_in_range")).define("calories_in_range", ["fitbit_range_call"], function(fitbit_range_call){return(
async (start, end) => fitbit_range_call(
  `https://api.fitbit.com/1/user/-/foods/log/caloriesIn/date/${start}/${end}.json`,
  "foods-log-caloriesIn")
)});
  main.variable(observer("calories_out_range")).define("calories_out_range", ["fitbit_range_call"], function(fitbit_range_call){return(
async (start, end) => fitbit_range_call(
  `https://api.fitbit.com/1/user/-/activities/calories/date/${start}/${end}.json`,
  "activities-calories")
)});
  main.variable(observer("weight_range")).define("weight_range", ["fitbit_range_call"], function(fitbit_range_call){return(
async (start, end) => fitbit_range_call(
  // NOTE1: data may miss days!
  // NOTE2: only works for range of at most 31 days
  `https://api.fitbit.com/1/user/-/body/log/weight/date/${start}/${end}.json`,
  "weight")
)});
  return main;
}
